package coil.geek.storage;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

/**
 * A disk based storage implemenation for {@link PersistentCacheLoader}
 * 
 * Copyright (c) 2009, Oded Arbel
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *   * Redistributions of source code must retain the above copyright notice, this 
 *     list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright notice, this 
 *     list of conditions and the following disclaimer in the documentation and/or 
 *     other materials provided with the distribution.
 *   * The names of its contributors may not be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY 
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 *   
 * @author Oded Arbel <oded@geek.co.il>
 *
 * @param <T>
 */
class DiskCacheLoader<T extends Serializable> extends PersistentCacheLoader<T> {
	private static String s_basePath = System.getProperty("coil.geek.storage.base-path", "/var/tmp/geek");
	private static long s_maxSize = Long.parseLong(System.getProperty("coil.geek.storage.max-size","100000000"));
	
	private File m_basePath;
	
	DiskCacheLoader(ObjectBackendLoader<T> backend) {
		super(backend);
		m_basePath = new File(s_basePath);
		if (!m_basePath.exists() && !m_basePath.mkdirs())
			throw new RuntimeException("Failed to initialize disk based cache loader: base path " + m_basePath + " cannot be created!");
		if (!m_basePath.isDirectory())
			throw new RuntimeException("Failed to initialize disk based cache loader: base path " + m_basePath + " is not a directory!");
	}

	private String flattenKey(String key) {
		String safe = key.replaceAll("\\\\","_");
		return safe.substring(0,1) + "/" + safe.substring(1,3) + "/" + safe.substring(3,6) + "/" + safe.substring(6);
	}
	
	private File generateFile(String key) {
		return new File(m_basePath.getAbsolutePath() + "/" + flattenKey(key));
	}

	public void store(String key, T object) throws IOException {
		File store = generateFile(key);
		File dir = store.getParentFile();
		if (!dir.exists() && !dir.mkdirs())
			throw new RuntimeException("Failed to create directory to store " + key + "!");
		File tmp = File.createTempFile("cache-store", ".tmp", m_basePath);
		ObjectOutputStream os = null;
		try {
			os = new ObjectOutputStream(new FileOutputStream(tmp));
			os.writeObject(object);
		} finally {
			if (os != null) {
				os.flush();
				os.close();
			}
			if (!tmp.renameTo(store))
				throw new IOException("Failed to rename temp file to target " + store);
		}
	}
	
	public boolean has(String key) throws IOException {
		File store = generateFile(key);
		return store.exists();
	}
	
	@SuppressWarnings("unchecked")
	public T load(String key) throws IOException {
		File store = generateFile(key);
		if (!store.exists())
			throw new RuntimeException("Can't find store file for " + key + "!");
		ObjectInputStream is = null;
		try {
			is = new ObjectInputStream(new FileInputStream(store));
			return (T)is.readObject();
		} catch (ClassNotFoundException e) {
			throw new IOException("Object loading failed as no class definition is available to the loaded object: " + e.getMessage());
		} finally {
			if (is != null)
				is.close();
		}
	}
	
	private void listFiles(File dir, List<File> filelist) {
		for (File p : dir.listFiles(new FilenameFilter(){
			public boolean accept(File dir, String name) {
				return !(name.startsWith("."));
			}
		})) {
			if (p.isDirectory())
				listFiles(p, filelist);
			else
				filelist.add(p);
		}
	}
	
	public void purge() {
		// remove everything over 100MB based on access time
		LinkedList<File> files = new LinkedList<File>();
		listFiles(m_basePath, files);
		Collections.sort(files, new Comparator<File>() {
			public int compare(File o1, File o2) {
				return (int) (o1.lastModified() - o2.lastModified());
			}});
		long sum = 0;
		for (File f : files)
			sum+= f.length();
		while (sum > s_maxSize) {
			File f = files.removeFirst();
			sum -= f.length();
			f.delete();
		}
	}
}

