package coil.geek.storage;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;

/**
 * A MySQL based storage implementation for {@link PersistentCacheLoader}
 * 
 * Copyright (c) 2009, Oded Arbel
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *   * Redistributions of source code must retain the above copyright notice, this 
 *     list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright notice, this 
 *     list of conditions and the following disclaimer in the documentation and/or 
 *     other materials provided with the distribution.
 *   * The names of its contributors may not be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY 
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 * @author Oded Arbel <oded@geek.co.il>
 *
 * @param <T> Type of objects to be persisted
 */
public class MysqlCacheLoader<T extends Serializable> extends PersistentCacheLoader<T> {
	private static String s_dbhost = System.getProperty("coil.geek.storage.host", "localhost");
	private static String s_dbname = System.getProperty("coil.geek.storage.database", "storage");
	private static String s_dbuser = System.getProperty("coil.geek.storage.username", "root");
	private static String s_dbpass = System.getProperty("coil.geek.storage.password", null);
	private static long s_maxSize = Long.parseLong(System.getProperty("coil.geek.storage.max-size","100000000"));
	
	private Connection m_database;

	MysqlCacheLoader(ObjectBackendLoader<T> backend) {
		super(backend);
		try {
			DriverManager.registerDriver(new com.mysql.jdbc.Driver());
			initConnection();
		} catch (SQLException e) {
			throw new RuntimeException("Failed to initialize MySQL cache loader: " + e);
		}
		
		try {
			executeQuery("SELECT COUNT(*) FROM object_table").close();
		} catch (SQLException e) {
			try {
				executeQuery("CREATE TABLE object_table (id INT PRIMARY KEY auto_increment, updated TIMESTAMP, name VARCHAR(255) UNIQUE, value LONGBLOB);");
			} catch (SQLException e1) {
				throw new RuntimeException("Failed to initialize MySQL cache loader: " + e);
			}
		}
	}
	
	private ResultSet executeQuery(String sql, Object ... data) throws SQLException {
		PreparedStatement st = m_database.prepareStatement(sql);
		for (int i = 1; i <= data.length; i++) {
			if (data[i-1] instanceof String)
				st.setString(i, (String) data[i-1]);
			else if (data[i-1] instanceof InputStream)
				st.setBlob(i, (InputStream) data[i-1]);
			else
				st.setObject(i, data[i-1]);
		}
		if (st.execute())
			return st.getResultSet();
		return null;
	}
	
	private void initConnection() throws SQLException {
		if (m_database != null)
			m_database.close();
		m_database = DriverManager.getConnection("jdbc:mysql://" + s_dbhost + "/" + s_dbname, s_dbuser, s_dbpass);
		m_database.setAutoCommit(true);
	}

	@Override
	public boolean has(String key) throws IOException {
		ResultSet s = null;
		try {
			s = executeQuery("SELECT COUNT(id) FROM object_table WHERE name = ?", key);
			if (s.next() && s.getInt(1) > 0)
				return true;
		} catch (SQLException e) {
			throw new IOException(e);
		} finally {
			if (s != null)
				try {
					s.close();
				} catch (SQLException e) {}
		}
		return false;
	}

	@SuppressWarnings("unchecked")
	@Override
	public T load(String key) throws IOException {
		ResultSet s = null;
		try {
			s = executeQuery("SELECT value FROM object_table WHERE name = ?", key);
			if (!s.next())
				throw new IOException("Can't load " + key + " as it doesnot exist in database");
				
			ObjectInputStream is = new ObjectInputStream(s.getBlob(1).getBinaryStream());
			return (T) is.readObject();
		} catch (SQLException e) {
			throw new IOException(e);
		} catch (ClassNotFoundException e) {
			throw new IOException("Object loading failed as not class definition is available to the loaded object: " + e.getMessage());
		} finally {
			if (s != null)
				try {
					s.close();
				} catch (SQLException e) {}
		}
	}

	@Override
	public void purge() throws IOException {
		ResultSet s = null;
		long sum = 0;
		LinkedList<Map.Entry<Integer,Long>> objects = new LinkedList<Map.Entry<Integer,Long>>();
		try {
			s = executeQuery("SELECT id, length(value) FROM object_table ORDER by updated ASC");
			while (s.next()) {
				final int id = s.getInt(1);
				final long len = s.getLong(2);
				objects.add(new Map.Entry<Integer, Long>() {
					public Integer getKey() { return id; }
					public Long getValue() { return len; }
					public Long setValue(Long value) { return len; }
				} );
				sum += len;
			}
			s.close();
			while (sum > s_maxSize) {
				Entry<Integer, Long> o = objects.removeFirst();
				sum -= o.getValue();
				executeQuery("DELETE FROM object_table WHERE id = ?", o.getKey());
			}
			executeQuery("OPTIMIZE TABLE object_table");
		} catch (SQLException e) {
			throw new IOException(e);
		} finally {
			try {
				if (s != null)
					s.close();
			} catch (SQLException e) {}
		}
	}

	@Override
	public void store(String key, final T object) throws IOException {
		PipedOutputStream pipeout = new PipedOutputStream();
		PipedInputStream pipein = new PipedInputStream(pipeout);
		final ObjectOutputStream os = new ObjectOutputStream(pipeout);
		PipedOutputStream pipeout2 = new PipedOutputStream();
		PipedInputStream pipein2 = new PipedInputStream(pipeout2);
		final ObjectOutputStream os2 = new ObjectOutputStream(pipeout2);
		new Thread() {
			@Override
			public void run() {
				try {
					os.writeObject(object);
					os.flush();
					os.close();
					os2.writeObject(object);
					os2.flush();
					os2.close();
				} catch (IOException e) {
					throw new RuntimeException("Failed to write object to database stream",e);
				}
			}
		}.start();
		try {
			executeQuery("INSERT INTO object_table (name, value) VALUES (?,?) ON DUPLICATE KEY UPDATE value = ?", key, pipein, pipein2);
		} catch (SQLException e) {
			throw new IOException(e);
		}
	}
}
