package coil.geek.storage;

import java.io.IOException;
import java.io.Serializable;
import java.util.Random;

/**
 * An implementation of a persistent object cache, backed by a user provided data
 * provider.
 * 
 * Usage: 
 *  - Use {@link #init(CacheType, ObjectBackendLoader)} to get a cache loader instance by 
 * providing your own implementation for handling cache misses.
 *  - Call {@link #save(String, Serializable)} to cache objects mapped to keys.
 *  - Call {@link #rescue(String)} to retrieve data from the cache or user backend.
 *  - You may optionally call {@link #purge()} periodically to keep the cache down to size
 * but this is optional: <tt>save()</tt> will do this for you automatically.
 * 
 *  The cache implementations use the following system properties to get configuration:
 *  coil.geek.storage.max-size : max size in bytes for the cache (default 100000000 bytes).
 *  coil.geek.storage.base-path : base directory for the disk cache (default /var/tmp/geek).
 *  coil.geek.storage.host : server name for the MySQL cache
 *  coil.geek.storage.database : database (schema) name for the MySQL cache
 *  coil.geek.storage.username : database login for the MySQL cache
 *  coil.geek.storage.password : database password for the MySQL cache
 * 
 * Copyright (c) 2009, Oded Arbel
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *   * Redistributions of source code must retain the above copyright notice, this 
 *     list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright notice, this 
 *     list of conditions and the following disclaimer in the documentation and/or 
 *     other materials provided with the distribution.
 *   * The names of its contributors may not be used to endorse or promote products 
 *     derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY 
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 * @author Oded Arbel <oded@geek.co.il>
 *
 * @param <T> Object type to be stored and retrieved.
 */
abstract class PersistentCacheLoader<T extends Serializable> {
	public static enum CacheType {
		DISK,
		MYSQL
	}

	private ObjectBackendLoader<T> m_backend;
	
	/**
	 * Private constructor. use {@link #init(CacheType, ObjectBackendLoader)} to get a new
	 * instance
	 * @param backend user provided data backend
	 */
	protected PersistentCacheLoader(ObjectBackendLoader<T> backend) {
		m_backend = backend;
	}

	/**
	 * Try to load an old object from the cache. If the data cannot be located, the user's 
	 * backend data provider will be called to get the object.
	 * @param key identifier for the object to load
	 * @return Object loaded from cache or retrieved fresh from the backend.
	 * @throws IOException In case of loading errors
	 */
	public T rescue(String key) throws IOException {
		if (has(key))
			return load(key);
		return m_backend.load(key);
	}
	
	/**
	 * Save an object to the cache.
	 * @param key identifier fro the object to save
	 * @param object data to save
	 * @throws IOException In case of storage error
	 */
	public void save(String key, T object) throws IOException {
		if (new Random().nextInt() % 1000 == 0)
			purge();
		store(key, object);
	}
	
	protected abstract void store(String key, T object) throws IOException;
	protected abstract boolean has(String key) throws IOException;
	protected abstract T load(String key) throws IOException;
	public abstract void purge() throws IOException;
	
	/**
	 * Create a new cache loader instance with the specific implementation requires
	 * @param <G> Type of object to be stored
	 * @param type cache implementation type - disk or MySQL based
	 * @param backend user provided backend to handle cache misses
	 * @return a new cache loader instance
	 */
	static public <G extends Serializable> PersistentCacheLoader<G> init(CacheType type, ObjectBackendLoader<G> backend) {
		switch (type) {
			case DISK: return new DiskCacheLoader<G>(backend);
			case MYSQL: return new MysqlCacheLoader<G>(backend);
		}	
		return null;
	}
}
